"""Basic specification for backend.

This file provides the basic interface that all backends shall implement.
See the Backend class for further information.
"""

import importlib

from oxtie.fronts import base as front_base


class Backend(object):
    """Class specifying main interface for backend and providing a few utils.

    All other backend implementations shall inherit from this and follow
    this model.
    """

    def __init__(self, front_cls=None, allow_load=False):
        """Initializer.

        :param front_cls=None:  Class to call to create an instance of an
                                oxtie front end. Calling
                                `front_cls.deserialize(raw, self)` shall
                                return an instance of
                                `oxtie.fronts.base.Frontend`. This
                                `front_cls` argument can be None (e.g.,
                                if `allow_load` is True or if you plan to
                                deal with this later).

        :param allow_load=False: Whether to allow dynamic import of the
                                 front_cls at load time. This is technically
                                 a security risk since you are dynamically
                                 loading a python module so it is disabled
                                 by default. Provided you trust the modules
                                 on your system or the backend you are loading
                                 from, though, it can be relatively safe.
        """
        self.front_cls = front_cls
        self.allow_load = allow_load

    def save(self, front):
        """Save the given instance to the backend.

        :param front:  Instance of `oxtie.fronts.base.Frontend` to save.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  This method saves the given front instance to the backend
                  so that calling self.load(front.name) will return something
                  matching front. Sub-classes shall implement this.

                  Generally sub-classes do not need to override this but
                  can instead override write_serialized instead.

        """
        serialized = front.serialize()
        self.write_serialized(front, serialized)

    def load(self, uid, only_hdr=False, front_cls=None, allow_load=None):
        """Load a saved instance from the backend.

        :param uid:    Either a string name or a Frontend instance or some
                       other unique id that the backend supports for loading
                       an item. A string or Frontend shall always be supported
                       while other forms of uid depend on the backend.

        :param only_hdr=False:  If True, only deserialize the header which
                                is faster and cheaper.

        :param front_cls=None:  The class we are deserializing. You can leave
                                this as None if only_hdr is True or if you
                                set allow_load to True (see below).

        :param allow_load=None: If this is None then use self.allow_load.
                                If allow_load evaluates to True, then we will
                                try to dynamically load the module and class
                                that the named object was serialized from.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  Instance of `oxtie.fronts.base.Frontend` which was
                  previously saved with the given name.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Load front end instance that was saved.

                  Sub-classes generally do not need to implement this
                  method. They can implement get_read_stream instead.

        """
        if only_hdr and front_cls is None:   # If only_hdr we can juse assume
            front_cls = front_base.Frontend  # generic Frontend for now.
        if front_cls is None:
            if allow_load or (allow_load is None and self.allow_load):
                hdr = self.load(uid, only_hdr=True)
                mod = importlib.import_module(hdr['mod_name'])
                front_cls = getattr(mod, hdr['cls_name'])
        stream = self.get_read_stream(uid)
        result = front_cls.deserialize(stream, self, only_hdr=only_hdr)
        return result

    def get_read_stream(self, uid):
        """Return a file-like object to read serialized representation of name.

        :param uid:    Either a string name or a Frontend instance or some
                       other unique id that the backend supports for loading
                       an item.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:   A file-like object with a read method.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Sub-classes can implement this method in order to allow
                   load to work.

        """
        raise NotImplementedError

    def write_serialized(self, front, serialized):
        """Write a serialized representation of front to this backend.

        :param front:    Instance of the thing we are serializing.

        :param serialized:  The serialized representation of front.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Sub-classes shall implement this method to write the
                   data to the backend.

        """
        raise NotImplementedError

    def delete(self, uid):
        """Delete the object with the given uid if present.

        :param uid:    Either a string name or a Frontend instance or some
                       other unique id that the backend supports for loading
                       an item.

        It is considered undefined to delete an object if it does not exist.
        Sub-classes can either raise KeyError or do nothing in such cases.
        This is because different backends may have different performance
        costs for verifying if an object exists before deleting it.
        """
        raise NotImplementedError



"""Specialized examples of frontends with more features.
"""

import os
import tempfile
import re
import doctest

from oxtie.fronts import nums
from oxtie.backs import aws, simple


class ColValidatorMixin(object):  # pylint: disable=too-few-public-methods
    """Mix-in to require certain columns in something with frame property.

    The ColValidatorMixin illustrates how to have a validate method
    which verifies that a frame property with certain columns exists.

    **IMPORTANT**: Your mixins must come before other things to get called.

    To illustrate, we do some setup to create an example pandas DataFrame:

>>> import pandas, io, logging
>>> from oxtie.fronts import specials, nums
>>> csv = ('%c'%10).join(['event_date,estimate,reported,period_date',
...     '2017-10-01,,,2017/9',  # Create some sample data
...     '2017-07-01,0.17,0.15,2017/6',''])
>>> logging.info('IMPORTANT: Mixins go before other classes to work right.')
>>> class RequireEstimateExample(specials.ColValidatorMixin, nums.SimpleFrame):
...     required_cols = set(['estimate']) # Set of col names we require.
...

    Next we deserialize from the CSV representation:

>>> f = RequireEstimateExample.deserialize_body_csv(
...       {'_name': 'test'}, {'__csv': csv}, None)
>>> f.frame.to_csv() == csv  # Verify deserialized version matches.
True

    Finally, we illustrate what happens if we try to deserialize from
    a version which is missing some columns. Note that this is kind of a nice
    way to build some validation into how you save and load data. Here we
    are just taking CSV data that may have been generated from who knows where
    and trying to deserialize it into an instance of our RequireEstimateExample
    class.

>>> try:
...     RequireEstimateExample.deserialize_body_csv({'_name': 'test'},
...        {'__csv': f.frame[['reported']].to_csv()},  # Only provide 1 column
...        None)
... except ValueError as exc:
...     print('Got ValueError: %s' % str(exc))
...
Got ValueError: Missing columns: [estimate]

    """

    # Set required_cols to a set of strings to check.
    required_cols = set()

    def validate(self):
        """Validate that required_cols are in self.frame
        """
        super().validate()
        frame = getattr(self, 'frame', None)
        if frame is None:
            raise ValueError('Missing columns %s since no frame' % ', '.join(
                self.required_cols))
        cols = set(list(self.frame))
        missing = sorted(self.required_cols - cols)
        if missing:
            raise ValueError('Missing columns: [%s]' % ', '.join(missing))


class ColRegexpMixin(object):  # pylint: disable=too-few-public-methods

    # Set required_cols to a set of strings to check.
    col_regexps = []

    def validate(self):
        """Validate that required_cols are in self.frame
        """
        super().validate()
        frame = getattr(self, 'frame', None)
        if frame is None:
            raise ValueError('Missing columns %s since no frame' % ', '.join(
                [c[0] for c in self.col_regexps]))
        for col_name, c_re in self.col_regexps:
            if col_name not in self.frame:
                raise ValueError('Missing required column %s' % col_name)
            if c_re is None:
                continue  # Just verified that column is present
            c_re_c = re.compile(c_re)
            for date, item in self.frame[col_name].iteritems():
                if not c_re_c.match(item):
                    raise ValueError(
                        'In column %s, index %s, item %s fails regexp %s' % (
                            col_name, date, item, c_re))


class S3BackendMixin(object):  # pylint: disable=too-few-public-methods
    """Backend mixin to store to S3.

    This class provides a mixin you can include in defining a frontend
    which will store to S3 by default. You can set the default_bucket
    class property to choose your bucket and the default_category class
    property to set a prefix `directory` in your S3 bucket if you want.
    """

    default_bucket = 'oxtie'  # Set to string indicating your S3 bucket.
    default_category = ''      # Set to string indicating a prefix (can be '')

    @classmethod
    def get_backend(cls, backend=None):
        """Create an S3Backend instance.
        """
        return backend if backend else aws.S3Backend(
            category=cls.default_category, bucket_name=cls.default_bucket)


class TempFileBackend(simple.FileBackend):
    """Backend to store to temporary file.

>>> import pandas  # So we can make a dataframe.
>>> from oxtie.fronts import specials  # Illustrate special example
>>> frame = pandas.DataFrame({'estimate': [.17, None]}, # Make some
...     index=['2017-07-01', '2017-10-01'])             # sample data
>>> backend=specials.TempFileBackend()
>>> f = specials.nums.SimpleFrame(name='test', backend=backend, frame=frame)
>>> f.save()
>>> g = backend.load('test', allow_load=True)
>>> g.frame.to_csv() == f.frame.to_csv()
True
    """

    def __init__(self, root=None, *args, **kwargs):
        if root is None:
            root = os.path.dirname(tempfile.mktemp())
        super().__init__(root, *args, **kwargs)


class EarningsFrame(ColValidatorMixin, S3BackendMixin, nums.SimpleFrame):
    """Example to show how you could save/load earnings data to/from S3.

    The EarningsFrame illustrates how to sub-class nums.SimpleFrame and
    use some simple mix-ins to create a nice object which can use a
    Pandas DataFrame to store stock earning data which is easy to save/load
    to/from Amazon S3.

    To illustrate, we do some setup to create an example pandas DataFrame:

>>> import pandas, io
>>> from oxtie.fronts import specials
>>> frame = pandas.DataFrame({'estimate': [.17, None], 'reported': [.15, None],
...     'period_date': ['2017/6', '2017/9']},
...     index=['2017-07-01', '2017-10-01'])
>>> f = specials.EarningsFrame('test', frame=frame)

    Now you would just call `f.save()` to save to s3.
    To load from s3, you would just call `specials.EarningsFrame.load('test')`.
    Note that you will need to have the S3 bucket specified in
    the `specials.EarningsFrame.default_bucket` class variable created and
    have your S3 credentials in ~/.aws/config for this to work.
    """

    default_bucket = 'oxtie'
    default_category = 'data/earnings'
    required_cols = set(['period_date', 'estimate', 'reported'])


if __name__ == '__main__':
    doctest.testmod()
    print('Finished tests')



"""Basic specification for a frontend.

See docs on Frontend class for further information.
"""

import doctest

import msgpack
import oxtie


class Frontend(object):
    """Class specifying the main interface for a frontend and some utils.

    All other frontend implementations shall inherit from this.

    The following illustrates example usage. First we do imports:

>>> import tempfile, os, shutil
>>> from oxtie.fronts import base
>>> from oxtie.backs import simple

    Next we create an instance of simple FileBackend and give it a
    root directory so it knows where to store things.

>>> backend = simple.FileBackend(root=tempfile.mkdtemp(suffix='.oxtie'))

    Now we crate an instance of our Frontend object with the name
    'test', some attributes, and tell it to use our backend.

>>> f = base.Frontend('test', {'last_update': '2017', 'bar': 'baz'}, backend)

    We could have sub-classed base.Frontend to have additional properties
    or we can just add them whenever we like:

>>> f.big_data = '*' * 10  # You can add properties and they will get saved

    Now we simply call the save method and the backend saves it. The nice
    thing here is that the f.save() could would work equally well with any
    backend. This gives you a lot of flexbility since you can choose or
    change your backend later.

>>> f.save()

    We can load just the header for our saved object if desired. This is
    usually very cheap and fast. The header also provides the attributes
    (which should be kept small) so you can load the header to decide if
    you want to do further processing:

>>> hdr = backend.load('test', only_hdr=True) # Fast load of just header
>>> hdr['_attributes']['last_update']
'2017'

    Imagine we do want to load the item. We can just call the backend.load
    method. Here we explicitly tell the backend what python class to load
    the data into. Later we can explore some clever tricks so that you
    don't need to specify the frontend because the backend can read it
    from the header and find the right class itself.

>>> g = backend.load('test', front_cls=base.Frontend)
>>> f.big_data == g.big_data  # Verify things worked
True

    If we wanted to explicitly lookup the frontend class, we could do:

>>> import importlib
>>> mod = importlib.import_module(hdr['mod_name'])
>>> h = backend.load('test', front_cls=getattr(mod, hdr['cls_name']))
>>> h.big_data == f.big_data
True

    If we are OK with implicitly loading modules we can just specify
    `allow_load=True` and things will be loaded implicitly. This is
    not done by default since loading arbitrary modules is potentially
    a security risk and it might get confusing if your paths are not set right.

>>> i = backend.load('test', allow_load=True)
>>> i.big_data == f.big_data
True

    Now cleanup the temporary directory we used for this demo.

>>> shutil.rmtree(backend.root)

    """

    def __init__(self, name, attributes, backend=None):
        """Initializer.

        :param name:   String name for this instance. This will be used as
                       a key by the backend so it should be unique.

        :param attributes:   Dictionary of attributes for the item. This can
                             be {} if you have no attriubutes. Ideally it
                             should be a dictionary of string keys and
                             simple types (e.g., int, float, string, dates)
                             as values. Which can be serialized with msgpack.
                             An import feature of the attributes is that
                             they are packed/unpacked in the header so you
                             can easily/quickly assess whether to do a full
                             unpack or not or scan packed items.

        :param backend=None: Instance of a backend to use in saving. This can
                             be None if you don't intend to save it or if you
                             want the get_backend method to choose it laer.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Main initializer for a Frontend. Sub-classes will almost
                  certainly have additional properties to store.
        """
        self._name = name
        self._attributes = attributes if attributes is not None else {}
        self._backend = backend

    def get_name(self):
        """Return the name of this object.

        Use get_name instead of accessing self._name directly for safety.
        """
        return self._name

    def get_attr_dict(self):
        """Return a reference to the self._attributes dict set in __init__.
        """
        return self._attributes

    def serialize(self, body_mode=None):
        """Serialize self into a string.

        :param mode=None:   Optional string indicating mode. If None, then
                            we use self.get_default_body_mode().

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  A serialized representation of self.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Workhorse to do the serialization before we save something.
                  You generally do **NOT** need to override this method in
                  a sub-class. Instead, you can implement a new mode and
                  provide a serialize_body_<MODE> method which this method will
                  lookup and call.

                  By default we always serialize the header with msgpack but
                  allow serializing the body in some other mode. This is so
                  that in deserializing we can always unpack the header using
                  msgpack and then decide what to do next.
        """
        if body_mode is None:
            body_mode = self.get_default_body_mode()
        hdr = self.serialize_header()
        bfunc = getattr(self, 'serialize_body_%s' % body_mode, None)
        if bfunc is None:
            raise ValueError('Invalid body serialization mode %s' % body_mode)

        body = bfunc()  # pylint: disable=not-callable
        return hdr + msgpack.packb(body, use_bin_type=True)

    def serialize_header(self):
        """Serialize the header.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  A seraizlied version of the header.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  We generally want to serialize the header and body
                  separately. This is because:

             1. We want the header in a simple standard format that is
                easily read. (We use msgpack by default). While there may
                be reasons to use different serialization for the body, it is
                essential that you can *READ* the header easily so you know
                what is going on with the body.
             2. You generally want the header to be short and quick so you
                can get a little information before deciding whether to
                read/parse the body.

        """
        header = {'body_mode': 'msgpack', '__oxtie_version__': oxtie.version,
                  '_name': self._name, '_attributes': self._attributes,
                  'mod_name': self.__module__,
                  'cls_name': self.__class__.__name__}
        return msgpack.packb(header, use_bin_type=True)

    def serialize_body_msgpack(self, do_pack=True, skips=()):
        """Serialize the body in msgpack format and return it.

        :param do_pack=True: If True, return body in msgpack format, otherwise
                             return the dictionary we would pack but leave it
                             as a dict in case the caller wants to modify it
                             before serializing. This is useful for
                             sub-classes.

        :param skips=(): Tuple of strings indicating properties to skip. We
                         always skip _backend, _name, and _attributes since
                         those eitehr go in the header or are not supposed
                         to be serialized. If you implement a sub-class that
                         serializes some properties in a special way, you can
                         pass in the name of properties you handle serparately
                         here.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  Bytes representing serialized version of the body of self.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Meant to be called by serialize method and not directly.

        The default implementation serializes everything in self.__dict__
        except for the given `skips` which are serialized in the header.
        If you have special properties, you may want to override this method
        to serialize your special items differntly.

        NOTE: If you do something special here, you may also need to
              override deserialize_body_msgpack as well.
        """
        skips = set(skips).union(('_backend', '_name', '_attributes'))
        body = {n: v for n, v in self.__dict__.items() if n not in skips}
        if do_pack:
            body = msgpack.packb(body, use_bin_type=True)
        return body

    @classmethod
    def get_default_body_mode(cls):
        """Return string indicating default serialization body mode.
        """
        return 'msgpack'

    @classmethod
    def deserialize(cls, packed, backend=None, body_mode=None, only_hdr=False):
        """Basically the opposite of serialize: make instance from data.

        :param packed:   The packed serialized representation produced by
                         the serialize method.

        :param backend:  The backend that you would like to use to be able
                         to store this instance. You can leave this as None
                         if you just want to use the item not store it again.

        :param body_mode=None:  Optional string mode as with serialize method.
                           If None, we first look at the body_mode in header
                           and if that is not use get_default_body_mode. You
                           can implement your own deserialize_body_<BODY_MODE>
                           method so you don't have to override this method.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:   Instance of this class matching the one which was
                   serialized into packed.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Inverse of serialize: produce an object from data.

                   Note: you generally do *NOT* want to override this but
                   instead override deserialize_body_* methods.

        """
        if only_hdr:
            return cls.packed_to_hdr(packed)
        hdr, body = cls.packed_to_hdr_and_body(packed)
        if body_mode is None:
            body_mode = hdr.get('body_mode', None)
        if body_mode is None:
            body_mode = cls.get_default_body_mode()
        myfunc = getattr(cls, 'deserialize_body_%s' % body_mode, None)
        if myfunc is None:
            raise ValueError('Invalid serialization body_mode %s' % body_mode)
        else:
            result = myfunc(hdr, body, backend)  # pylint: disable=not-callable
            result.validate()
            return result

    @classmethod
    def packed_to_hdr_and_body(cls, packed):
        """Unpack the header and body from a packed serialized representation.

        :param packed:     The packed serialized representation.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  The pair (hdr, body) where hdr is a dict containing
                  header information about the stored item and body is
                  the (still serialized) body.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Unpack the header and get the body. You will probably
                  still need to deserialize the body. This is a helper
                  method and not meant to be called directly.

        """
        dummy = cls
        unpacker = msgpack.Unpacker(packed, encoding='utf-8')
        hdr, body = unpacker
        return hdr, body

    @classmethod
    def packed_to_hdr(cls, packed):
        """Unpack the header from a packed serialized representation.

        :param packed:     The packed serialized representation.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  A dict representing the header information.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Quickly get just the header from the packed representation
                  so we can decide to do next.

        """
        unpacker = msgpack.Unpacker(packed, encoding='utf-8')
        return unpacker.unpack()

    @classmethod
    def deserialize_body_msgpack(cls, hdr, body, backend, skips=None):
        """Deserialize and create instance from packed version.

        :param hdr:       Dictionary for the header of the object.

        :param body:      Either Bytes representing msgpack representation of
                          the body or a dict. The latter is sometimes useful
                          for sub-classes which unpack some things themselves.

        :param backend:   Instance of backend to deserialize.

        :param skips=None:  Optional dictionary with string keys. Any key
                            in skips will *NOT* be deserialized into the
                            result but have its data placed into skips for
                            the caller to handle. This is useful for
                            sub-classes which are overriding this.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

         :return:   Instance of cls.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Inverse of serialize.  Meant to be called by deserialize
                   method when mode='msgpack'. Not meant to be called directly.

                   The default implementation takes the values in body
                   and puts them into the __dict__ for the new item.

        """
        skips = skips if skips else {}
        result = cls(hdr['_name'], hdr.get('_attributes', {}), backend)
        if isinstance(body, dict):
            my_dict = body
        else:
            my_dict = msgpack.unpackb(body, encoding='utf-8')
        for name, value in my_dict.items():
            if name in skips:
                skips[name] = value
            else:
                setattr(result, name, value)

        return result

    def validate(self):
        """Simple function to make sure self is valid.

        Should return True if valid or raise a ValueError otherwise.
        This is useful as a check to make sure things are deserialized
        correctly.
        """
        dummy = self
        return True

    @classmethod
    def get_backend(cls, backend=None):
        """Choose a backend instance.

        :param backend=None:  If provided, use this.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  Return backend if it is not None, otherwise try
                  to choose a reasonable bakend somehow.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Sub-clases can override this to provide a default backend.
        """
        return backend

    def save(self, backend=None):
        """Save an instance of self on the specified backend.

        :param backend=None:   Optional backend to use. If this is None,
                               we try to use self._backend (which is the
                               backend specified in __init__) and if that
                               fails we call self.get_backend().
        """
        backend = backend if backend else self.get_backend(self._backend)
        backend.save(self)

    @classmethod
    def load(cls, uid, backend=None, only_hdr=False):
        """Load an instance of self with the given name from specified backend.

        :param uid:    Either a string name or a Frontend instance or some
                       other unique id that the backend supports for loading
                       an item.

        :param backend=None:   The backend we want to load from. If this is
                               None, we call get_backend() to try to get one.

        :param only_hdr=False: Whether to only load the header (in which case
                               we do not need a backend).

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:   A new instance of self corresponding to the one we
                   `save`-ed previously if only_hdr is False or a dictionary
                   representing the header if only_hdr is True.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Inverse of save method; load it back.

        """
        backend = cls.get_backend(backend)
        new_item = backend.load(uid, only_hdr=only_hdr, front_cls=cls)
        return new_item

    def delete(self, backend=None):
        """Delete self from the backend.

        :param backend=None:   The backend we want to delete from. If this is
                               None, we call get_backend() to try to get one.

        The current instance is not changed; only backend representation is
        deleted. This should only be called if the backend actually is
        storing self.
        """
        backend = backend if backend else self.get_backend(self._backend)
        backend.delete(self)

if __name__ == '__main__':
    doctest.testmod()
    print('Finished Tests')



"""Backend to write to various Amazon Web Services (AWS) backends.

See the S3Backend and DynamoBackend classes.
"""

import io
import logging

import boto3
import botocore

from oxtie.backs import base as back_base


class S3Backend(back_base.Backend):
    """Backend for Amazon S3.

    See docs for __init__.
    """

    def __init__(self, category, bucket_name, *args, **kwargs):
        """Initializer.

        :param category:  Optional string prefix to use as `directory`. You
                          can leave this as '' to not use any prefix or
                          provide something like 'my/path/to/stuff' which
                          will be prepended like a directory to the name
                          of something to save.

        :param bucket_name:  AWS bucket name. You must have already created
                             this and have your ~/.aws/config set to give
                             permissions.

        :param *args, **kwargs:    As for base.Backend constructor (e.g.,
                                   front_cls, allow_load, etc.)

        """
        super().__init__(*args, **kwargs)
        self.category = category
        self.bucket_name = bucket_name
        self.my_s3 = boto3.session.Session().resource('s3')

    def prep_key(self, uid):
        """Convert a uid to key to use in saving/loading data.

        This can take a string name or a Frontend instance and call get_name().
        """
        if isinstance(uid, str):
            name = uid
        else:
            name = uid.get_name()

        if self.category:
            key = '%s/%s' % (self.category, name)
        else:
            key = name

        return key

    def write_serialized(self, front, serialized):
        """Implement as required by parent Backend clas.
        """
        key = self.prep_key(front)
        result = self.my_s3.Bucket(self.bucket_name).put_object(
            Key=key, Body=serialized)
        return result

    def get_read_stream(self, uid):
        """Implement as required by parent Backend class.
        """
        key = self.prep_key(uid)
        try:
            obj = self.my_s3.Object(self.bucket_name, key)
            raw = obj.get()['Body']
        except botocore.exceptions.ClientError as problem:
            logging.info('Got error in s3 operation: %s', str(problem))
            response = getattr(problem, 'response', None)
            if response is None:
                raise
            code = response.get('Error', {}).get('Code', None)
            if code is None:
                raise
            if code == 'NoSuchKey':
                raise KeyError(key)
            raise

        return raw

    def delete(self, uid):
        """Implement as required by parent Backend class.
        """
        key = self.prep_key(uid)
        self.my_s3.Object(self.bucket_name, key).delete()


class DynamoBackend(back_base.Backend):
    """Backend for saving data to AWS DynamoDB.

    See docs for __init__ for details.
    """

    default_key_name = None
    default_range_key = None
    default_range_value = None

    def __init__(self, table_name=None, key_name=None, range_key=None,
                 *args, **kwargs):
        """Initializer.

        :param table_name=None:  String name of dyanmodb table to use. This
                                 must already exists and you must have
                                 ~/.aws/config set to provide permissions.
                                 If None, we use self.default_table_name

        :param key_name=None:    Name of the primary key field. We use
                                 self.default_range_key if key_name is None.

        :param range_key=None:   Optional name of the secondary key field.
                                 This is a little tricky since most Backend
                                 sub-classes do not use secondary keys. If
                                 your table does have range keys, you would
                                 specify the field name of the range key here.
                                 When saving a Frontend instance, we look for
                                 get_attr_dict()[range_key] to get the range
                                 key. See also the docs for get_read_stream
                                 for loading something with a range_key.

        :param *args, **kwargs:    As for base.Backend constructor (e.g.,
                                   front_cls, allow_load, etc.)
        """
        super().__init__(*args, **kwargs)
        self.key_name = key_name if key_name is not None else (
            self.default_key_name)
        self.range_key = range_key if range_key is not None else (
            self.default_range_key)
        self.table_name = table_name if table_name is not None else (
            self.default_table_name)
        # Need session to keep things semi thread safe
        session = boto3.session.Session()
        self.dynamodb = session.resource('dynamodb')

    def prep_key(self, uid):
        """Convert a uid to key to use in saving/loading data.

        This can take a string name or a Frontend instance and call get_name().

        For this class, the uid can also be a pair of the form

                (HASH_KEY, RANGE_KEY)

        since if the table supports range keys, you will need to specify both.
        """
        if isinstance(uid, str):
            key = {self.key_name: uid}
            if self.range_key is not None:
                key[self.range_key] = self.default_range_value
        elif isinstance(uid, (tuple, list)):
            assert self.range_key is not None, (
                'No range_key specified in __init__')
            assert len(uid) == 2, (
                'Expected (hash_key, range_key) got %s' % (str(uid)))
            key = {self.key_name: uid[0],
                   self.range_key: uid[1]}
        else:
            key = {self.key_name: uid.get_name()}
            if self.range_key is not None:
                key[self.range_key] = uid.get_attr_dict().get(
                    self.range_key, self.default_range_value)
        return key

    def write_serialized(self, front, serialized):
        """Implement as required by parent Backend clas.
        """
        table = self.dynamodb.Table(self.table_name)
        data = self.prep_key(front)
        data['object'] = serialized
        table.put_item(Item=data)

    def get_read_stream(self, uid):
        """Implement as required by parent Backend clas.

        This can take a string name or a Frontend instance and call get_name().

        For this class, the uid can also be a pair of the form

                (HASH_KEY, RANGE_KEY)

        since if the table supports range keys, you will need to specify both.
        Thus when calling load(uid) you can specify (HASH_KEY, RANGE_KEY)
        to do the load.
        """
        table = self.dynamodb.Table(self.table_name)
        key = self.prep_key(uid)
        my_obj = table.get_item(Key=key)
        if my_obj is None or 'Item' not in my_obj:
            raise KeyError(key)
        result = io.BytesIO(my_obj['Item']['object'].value)
        return result

    def delete(self, uid):
        """Implement as required by parent Backend clas.
        """
        table = self.dynamodb.Table(self.table_name)
        key = self.prep_key(uid)
        table.delete_item(Key=key)



"""Simple implementations of basic storage methods for numeric data.

See the SimpleFrame class and its documentation to get started.
"""

import doctest
import io

import msgpack
import numpy
import pandas

from oxtie.fronts import base


def msgpack_encoder(obj):
    """Simple function to help with msgpack serialization.

    :param obj:    Object to serialize.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:    Either obj or a converted version which can be serialized.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:    Handle serializaing tricky things like numpy.bool_

    """
    if isinstance(obj, numpy.bool_):
        return bool(obj)
    return obj


class SimpleFrame(base.Frontend):
    """Storage of a pandas DataFrame

    The SimpleFrame class provides an oxtie front end to hold
    a pandas DataFrame. As with all front ends, you instantiate it
    with a name along with optional attributes and an optional backend.
    In addition, you can provide a frame which is a pandas DataFrame.

    The SimpleFrame knows how to intelligently store pandas DataFrame
    objects on a backend. It does things like handle specials types
    like numpy.bool_, Pandas Timestamp objects, etc.

    The following illustrates example usage:

    First we import various things and setup our DataFrame.

>>> import tempfile, os, shutil, pandas
>>> from oxtie.fronts import nums
>>> frame = pandas.DataFrame({'estimate': [.17, None], 'reported': [.15, None],
...     'period_date': ['2017/6', '2017/9']},
...     index=['2017-07-01', '2017-10-01'])

    Next we specify setup a backend that can save to a local file.

>>> from oxtie.backs import simple
>>> backend = simple.FileBackend(root=tempfile.mkdtemp(suffix='.oxtie'))

    Now we can create SimpleFrame instance.

>>> f = nums.SimpleFrame('test', backend=backend, frame=frame)
>>> f.more = '*' * 10  # Add some more data in a property to save.

    When we are ready, we can save the SimpleFrame frame.

>>> f.save()

    Later, we can reload the SimpleFrame simply by using its name.

>>> g = nums.SimpleFrame.load('test', backend=backend)

    Finally we verify that the loaded data matched the original and
    then cleanup.

>>> all(g.frame == f.frame) and f.more == g.more
True
>>> f.delete() # delete f from the backend.
>>> os.listdir(backend.root) # verify it was deleted.
[]
>>> shutil.rmtree(backend.root) # remove the temp directory we were using.
    """

    def __init__(self, name, attributes=None, backend=None, frame=None):
        """Initializer.

        :param name:  String name as for base.Frontend.

        :param attributes=None:   Attribtute dict as for base.Frontend.

        :param backend=None:    Optional backend as for base.Frontend.

        :param frame=None:    Optional pandas DataFrame.

        """
        super().__init__(name, attributes, backend)
        self.frame = frame

    def help_serialize_body(self):
        """Serialization helper.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  Data to serialize data frame for self in dictionary form.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  This method basically flattens self.frame into things
                  like the columns in the dataframe in self.frame, the
                  index, etc.  This is helpful since most
                  serialization systems are easier to apply to the
                  flattened pieces. Thus the way to write a new
                  serialize_body_* method is usually to first call
                  help_serial and then serialize the resulting
                  dictionary.

        SEE ALSO: help_deserialize_body

        """
        result = {'__columns': list(self.frame), '__index': []}
        idx = result['__index']
        result['__data'] = {n: [] for n in list(self.frame)}
        data_cols = [result['__data'][n] for n in list(self.frame)]
        my_tz = self._attributes.get('timezone', None)
        for item in self.frame.itertuples():
            idx_item = item[0]
            if my_tz is not None and hasattr(idx_item, 'tz_localize'):
                idx_item = idx_item.tz_localize(my_tz)
            if not isinstance(idx_item, str):
                idx_item = idx_item.timestamp()
            idx.append(idx_item)
            for dcol, value in zip(data_cols, item[1:]):
                dcol.append(value)
        return result

    def serialize_body_msgpack(self, do_pack=True, skips=None):
        """Implement serialization using msgpack protocol.

        This is meant to be called via serialize('msgpack') not directly. It
        overrides the serialize_body_msgpack to handle the pandas DataFrame
        intelligently.
        """
        if skips:
            skips = set(skips).union(('frame', ))
        else:
            skips = ('frame', )
        body = super().serialize_body_msgpack(do_pack=0, skips=skips)
        frame_body = self.help_serialize_body()
        assert not set(body).intersection(frame_body), (
            'Conflict in names for serializing body and frame')
        body.update(frame_body)
        if do_pack:
            body = msgpack.packb(body, use_bin_type=True,
                                 default=msgpack_encoder)
        return body

    @classmethod
    def deserialize_body_msgpack(cls, hdr, body, backend, skips=None):
        """Deserialize and create instance from packed version.

        Implements deserialize_body_msgpack as described in parent
        class with special handling for data frame.
        """
        skips = skips if skips else {}
        for name in ['__columns', '__index', '__data']:
            if name in skips:
                raise ValueError('Cannot skip %s in restoring frame' % name)
            skips[name] = None
        basic = super().deserialize_body_msgpack(hdr, body, backend, skips)
        return cls.help_deserial(basic, skips)

    @classmethod
    def deserialize_body_csv(cls, hdr, body, backend, skips=None):
        """Deserialize and create instance from packed CSV version.

        Implements deserialize_body_csv similarly to deserialize_body_msgpack.
        """
        skips = skips if skips else {}
        for name in ['__csv']:
            if name in skips:
                raise ValueError('Cannot skip %s in restoring frame' % name)
            skips[name] = None
        basic = super().deserialize_body_msgpack(hdr, body, backend, skips)
        wrap = io.StringIO(skips['__csv'])
        frame = pandas.read_csv(
            wrap, parse_dates=[0], infer_datetime_format=True, index_col=0)
        basic.frame = frame
        basic.validate()
        return basic

    @classmethod
    def help_deserial(cls, basic, skips):
        """Helper to deserialize.

        :param basic:     The basic object produced by the initial
                          deserialization step but with the frame not yet

        :param skips:     Dictionary with data elements that were skipped
                          in the basic deserialization and should be processed.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:   Fully deserialized instance.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   This takes a basic object which is mostly deserialized
                   except for the pands DataFrame and handles the DataFrame.
                   This lets us split the DataFrame handling code from
                   the generic deserialization code.

        """
        idx = cls.deserialize_timestamps(
            skips['__index'], basic.get_attr_dict().get('timezone'))
        fdict = {}
        for cname in skips['__columns']:
            cval = skips['__data'][cname]
            if cval:
                fdict[cname] = cval
            else:
                fdict[cname] = [None]*len(idx)

        # Explicitly specify columns so we get ordering as desired.
        frame = pandas.DataFrame(fdict, index=idx, columns=skips['__columns'])
        basic.frame = frame
        basic.validate()
        return basic

    @classmethod
    def deserialize_timestamps(cls, data, my_tz=None):
        idx = []
        for idx_item in data:
            if isinstance(idx_item, float):
                idx_item = pandas.Timestamp.fromtimestamp(idx_item)
            else:
                idx_item = pandas.Timestamp(idx_item)
            if my_tz is not None:
                idx_item = idx_item.tz_localize(my_tz)
            idx.append(idx_item)
        return idx

    @staticmethod
    def _regr_test():
        """
>>> import tempfile, os, shutil, pandas, random
>>> from oxtie.fronts import nums
>>> from oxtie.backs import simple
>>> random.seed(123)  # Make tests repeatable
>>> backend = simple.FileBackend(root=tempfile.mkdtemp(suffix='.oxtie'))
>>> idx = [pandas.Timestamp(
...     '2017-01-%s' % str(i).zfill(2)) for i in range(1,15)]
>>> estimate = [random.gauss(0, 1) for i in idx]
>>> reported = [e + .05 * random.gauss(0, 1) for e in estimate]
>>> p = pandas.DataFrame({'estimate': estimate, 'reported': reported},
...     index=idx)
>>> f = nums.SimpleFrame('test', {}, backend, p)
>>> f.more = '*' * 10  # Add some more data in a property to save.
>>> f.save()
>>> g = backend.load('test', allow_load=True)
>>> all(g.frame == f.frame) and f.more == g.more
True
>>> shutil.rmtree(backend.root)
"""

    @staticmethod
    def _regr_test_fromcsv():
        """Test reading in from a CSV since that has a few differences.

    Using pandas.read_csv produces a DataFrame with the following potential
    issues:

       1. Booleans become numpy.bool_ which may not be serializable without
          some tricks.
       2. Column ordering may end up different.

>>> import tempfile, os, shutil, io, pandas, msgpack
>>> from oxtie.fronts import nums
>>> from oxtie.backs import simple
>>> backend = simple.FileBackend(root=tempfile.mkdtemp(suffix='.oxtie'))
>>> data = io.StringIO(('%c'%10).join([
...    'event_date,estimate,period_date,reported,confirmed',
...    '10/1/2017,,2017/9,,FALSE',
...    '7/1/2017,0.17,2017/6,0.15,TRUE']))
>>> p = pandas.read_csv(data, index_col=0, parse_dates=[0])
>>> f = nums.SimpleFrame('test', {}, backend, p)
>>> f.more = '*' * 10  # Add some more data in a property to save.
>>> f.save()
>>> g = backend.load('test', allow_load=True)
>>> all(g.frame == f.frame) and f.more == g.more
True

    As an aside, you can use the deserialize_body_csv method to create
    a nums.SimpleFrame instance directly from a CSV representation without
    having to call pandas.read_csv first:

>>> cbody = {'__csv': f.frame.to_csv()}
>>> h = nums.SimpleFrame.deserialize_body_csv({'_name': 'test'}, cbody, None)
>>> all(h.frame == f.frame)
True

    Now clean up

>>> shutil.rmtree(backend.root)
"""

if __name__ == '__main__':
    doctest.testmod()
    print('Finished Tests')



"""This module provides various simple implementations of backends.
"""

import os

from oxtie.backs import base as back_base


class FileBackend(back_base.Backend):
    """File based backend.

    As described in __init__, you just need to provide a string `root`
    indicating the root directory. Then files will be stored inside
    that directory.

    """

    def __init__(self, root=None, *args, **kwargs):
        """Initializer.

        :param root=None:  String indicating root directory to save in.

        :param *args, **kwargs:  Passed to parent __init__. See docs
                                 for base.Backend class for details.

        """
        super().__init__(*args, **kwargs)
        self.root = root

    def write_serialized(self, front, serialized):
        """Implement as required by parent to write to file.
        """
        path = os.path.join(self.root, front.get_name())
        open(path, 'wb').write(serialized)

    def get_read_stream(self, uid):
        """Implement as required by parent to read from file.
        """
        path = os.path.join(self.root, self.prep_key(uid))
        stream = open(path, 'rb')
        return stream

    def prep_key(self, uid):
        """Prepare and return data key (i.e., file path) from uid.
        """
        dummy = self
        if isinstance(uid, str):
            return uid
        return uid.get_name()

    def delete(self, uid):
        path = os.path.join(self.root, self.prep_key(uid))
        os.remove(path)



"""oxtie package for storing data intelligently

The following are some useful places to look for more information:

  * oxtie.fronts: Package with implementations of front ends.
  * oxtie.backs:  Package with implementations of back ends.

"""

version = '0.1.1'
